using System;
using System.Collections.Generic;
using System.Text;
using System.ComponentModel;

namespace GameEngine {
	public class Grid : INotifyPropertyChanged {
		public Grid( ) {
			rowCount = 7;
			sequenceLengthForWin = 4;
		
			columns = new List<Column>( );
			for (int i = 1; i <= 7; i++)
				columns.Add(new Column(this));
			winner = null;

			nextPlayer = GetRandomDiscType( );
			
			// Only for visualization test purposes! Will kill all the tests!
			// RandomInit( );
		}

		public Grid(DiscType firstPlayer) : this() {
			this.nextPlayer = firstPlayer;
		}

		private DiscType GetRandomDiscType( ) {
			Random random = new Random( );
			return GetRandomDiscType(random);
		}

		private DiscType GetRandomDiscType(Random random) {
			return random.Next(2) == 1 ? DiscType.Red : DiscType.Yellow;
		}

		private void RandomInit( ) {
			Random random = new Random( );
			for (int col = 0; col <= 6; col++) {
				int discCount = random.Next(rowCount + 1);
				for (int i = 0; i < discCount; i++)
					Columns[col].AcceptDisc(GetRandomDiscType(random));
			}
		}

		private int rowCount;
		public int RowCount {
			get {
				return rowCount;
			}
			set {
				rowCount = value;
			}
		}

		private int sequenceLengthForWin;
		public int SequenceLengthForWin {
			get {
				return sequenceLengthForWin;
			}
			set {
				if (value > Math.Min(rowCount, 7))
					throw new ArgumentException("SequenceLengthForWin can't be greater than Min(RowCount, 7).");
				if (value < 2)
					throw new ArgumentException("SequenceLengthForWin can't be smaller than 2.");
				sequenceLengthForWin = value;
				CalculateWinner( );
			}
		}

		private List<Column> columns;
		public List<Column> Columns {
			get {
				return columns;
			}
		}

		private DiscType nextPlayer;
		public DiscType NextPlayer {
			get {
				return nextPlayer;
			}
		}

		public void AcceptMove(DiscType player, int column) {
			if (nextPlayer != player)
				throw new InvalidOperationException("Wrong player.");
			Columns[column].AcceptDisc(player);
			CalculateWinner( );
			nextPlayer = nextPlayer == DiscType.Red ? DiscType.Yellow : DiscType.Red;
			OnPropertyChanged("NextPlayer");
		}

		private DiscType? winner;
		public DiscType? Winner {
			get {
				return winner;
			}
		}

		private void CalculateWinner( ) {
			winner = FindVerticalWinner( );
			if (winner == null) {
				winner = FindHorizontalWinner( );
				if (winner == null)
					winner = FindDiagonalWinner( );
			}
			OnPropertyChanged("Winner");
		}

		private DiscType? FindVerticalWinner( ) {
			foreach (Column column in Columns) {
				DiscType? winner = column.FindWinner( );
				if (winner != null)
					return winner;
			}
			return null;
		}

		private DiscType? FindHorizontalWinner( ) {
			for (int row = 0; row < RowCount; row++) {
				DiscSequenceCounter sequenceCounter = new DiscSequenceCounter( );
				foreach (Column column in Columns) {
					if (row >= column.Discs.Count) {
						sequenceCounter.ResetSequence( );
						continue;
					}
					sequenceCounter.AcceptDisc(column.Discs[row]);
				}
				if (sequenceCounter.BestSequenceLength >= SequenceLengthForWin)
					return sequenceCounter.BestSequence;
			}
			return null;
		}

		private DiscType? FindDiagonalWinner( ) {
			// Start by checking the NE diagonals - i.e. these that run in NE direction.
			// First we need to go through the possible starting points.
			for (int row = RowCount - SequenceLengthForWin; row >= 0; row--) {
				DiscType? result = FindNEDiagonalWinner(0, row);
				if (result != null)
					return result;
			}
			for (int col = 0; col <= Columns.Count - SequenceLengthForWin; col++) {
				DiscType? result = FindNEDiagonalWinner(col, 0);
				if (result != null)
					return result;
			}

			// Now for the SE ones - same game.
			for (int row = SequenceLengthForWin - 1; row < RowCount; row++) {
				DiscType? result = FindSEDiagonalWinner(0, row);
				if (result != null)
					return result;
			}
			for (int col = 0; col < Columns.Count - SequenceLengthForWin; col++) {
				DiscType? result = FindSEDiagonalWinner(col, 0);
				if (result != null)
					return result;
			}
			return null;
		}

		private DiscType? FindSEDiagonalWinner(int col, int row) {
			DiscSequenceCounter sequenceCounter = new DiscSequenceCounter( );
			while (col < Columns.Count && row >= 0) {
				if (Columns[col].Discs.Count > row) 
					sequenceCounter.AcceptDisc(Columns[col].Discs[row]);
				else 
					sequenceCounter.ResetSequence();

				col++;
				row--;
			}
			return sequenceCounter.BestSequenceLength >= SequenceLengthForWin ?
				sequenceCounter.BestSequence : null;
		}

		private DiscType? FindNEDiagonalWinner(int col, int row) {
			DiscSequenceCounter sequenceCounter = new DiscSequenceCounter( );
			while (col < Columns.Count && row < RowCount) {
				if (Columns[col].Discs.Count > row)
					sequenceCounter.AcceptDisc(Columns[col].Discs[row]);
				else
					sequenceCounter.ResetSequence( );

				col++;
				row++;
			}
			return sequenceCounter.BestSequenceLength >= SequenceLengthForWin ?
				sequenceCounter.BestSequence : null;
		}

		#region INotifyPropertyChanged Members

		PropertyChangedEventHandler propertyChanged;
		protected virtual void OnPropertyChanged(string propertyName) {
			if (propertyChanged != null)
				propertyChanged(this, new PropertyChangedEventArgs(propertyName));
		}
		event PropertyChangedEventHandler INotifyPropertyChanged.PropertyChanged {
			add { propertyChanged += value; }
			remove { propertyChanged -= value; }
		}

		#endregion
	}
}